/*****************************************************************************
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/
/*                                                                           */
/*                                   #                                       */
/*                                   #                                       */
/*                                  ###                                      */
/*                                  ###                                      */
/*                                 #####                                     */
/*                                 #####                                     */
/*                                #######                                    */
/*                                #######                                    */
/*                   #################################                       */
/*                       #########################                           */
/*                           #################                               */
/*                  [[[[The Working Class Unlimited]]]                       */
/*                             #############                                 */
/*                             #############                                 */
/*                            ######   ######                                */
/*                            ###         ####                               */
/*                           ##              ##                              */
/*                          #                  #                             */
/*                                                                           */
/*                 (C) The Working Class Unlimited 1993.                     */

/**
 **
 **
 **                X p g p  -  P r i m i t i v e s   v2.27
 **
 **
 **
 **                                                   Milanovic Ljubomir, 1995.
 **/
#define NeedFunctionPrototypes 1

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <math.h>

#ifndef XPGP_XFONT
#define XPGP_XFONT "-*-times-medium-r-normal--%d-*-*-*-*-*-*-*"
#endif
#ifndef XPGP_PSFONT
#define XPGP_PSFONT "/Times-Roman"
#endif

#define Xpgp_NOD  -1				/* not output device */
#define Xpgp_X     0				/* X11 */
#define Xpgp_PS    1				/* PostScript */
#define Xpgp_BMP   2				/* BMP */

#define dtoi(x)  ((int)((x)+0.5))	        /* double to int (rounded) */


struct point {
    double x;
    double y;
};


struct __Xpgp_X_OutputDevice {			/* X output device */
    Display     *display;
    Window      window;
    Colormap    colormap;
    GC          gc;
    
    double      QPd;				/*quasi pixel parameters*/
    int         QPshape;
    
    unsigned long  *inks;			/*allocated colors*/
    int            NofInks;			/*number of allocated colors*/

    struct point gcur;				/* graph. cursor */
};


struct __Xpgp_PS_OutputDevice {			/* PostScript output dev. */
    FILE *fp;					/* output file pointer */
    int stroke_counter;				/* stroke counter */
    unsigned long sr_flag;			/* subrout. flag (circle,..) */

    int         orient;				/* orientation */
    double      QPd;				/* quasi pixel parameters */
    int         QPshape;
    struct inks {				/* allocated colors */
	unsigned short r;
	unsigned short g;
	unsigned short b;
    } *inks;
    int            NofInks;			/* number of allocated colors*/
    int            curPen;			/* current pen */
    struct point gcur;				/* graph. cursor */
    int            pageCounter;			/* page counter ( Clear() ) */
};

struct __Xpgp_BMP_OutputDevice {	        /* BMP output device */
    FILE *fp;					/* output file pointer */
    int  w, h;				        /* width, height */
    int  *bmp;				        /* bitmap */
    double      QPd;				/* quasi pixel parameters */
    int         QPshape;
    struct {				/* allocated colors */
	unsigned short r;
	unsigned short g;
	unsigned short b;
    } *inks;
    int            NofInks;			/* number of allocated colors*/
    int            curPen;			/* current pen */
    struct point gcur;				/* graph. cursor */
    int          lw;			        /* line width */
};



union __Xpgp_OutputDevice {			/* output device */
    struct __Xpgp_X_OutputDevice    X;
    struct __Xpgp_PS_OutputDevice   PS;
    struct __Xpgp_BMP_OutputDevice  BMP;
};

union __Xpgp_OutputDevice  *CurOD;		/* current output device */
int CurODtype;					/* current output dev. type */



/*
 * #define QPrect 0
 * #define QPcirc 1
 */

int __XpgpX_InitX11(int width,int height);
int __XpgpX_InitX11X(int width, int height, int xposition, int yposition,
      		     char* windowname, char* iconname, char* iconfile,
      		     char* icondata);

int __XpgpX_Clear(void);
int __XpgpX_fShow(void);

int __XpgpX_GetMouseStatus(int *x, int *y, unsigned int *status);
int __XpgpX_GetMousePressEvent(int *x, int *y, unsigned int *status);
int __XpgpX_SetFocus(union __Xpgp_OutputDevice *dev);

int __XpgpX_Ink(unsigned short red,unsigned short green,unsigned short blue);
int __XpgpX_ClearPalette(void);
int __XpgpX_ResetPalette(void);

int __XpgpX_SetBackground(int color);
int __XpgpX_Pen(int color);
int __XpgpX_Paper(int color);

int __XpgpX_SetLineWidth(double w);
int __XpgpX_SetDash(double b1, double w1, double b2, double w2);
int __XpgpX_SetFontSize(double size);
int __XpgpX_SetQP(double d, int shape);

int __XpgpX_Move(double x, double y);
int __XpgpX_MoveR(double dx, double dy);

int __XpgpX_Plot(double x, double y);
int __XpgpX_PlotR(double dx, double dy);
int __XpgpX_fPlot(double x, double y);
int __XpgpX_fPlotR(double dx, double dy);

int __XpgpX_Line(double x1, double y1, double x2, double y2);
int __XpgpX_LineTo(double x, double y);
int __XpgpX_LineR(double dx, double dy);
int __XpgpX_fLine(double x1, double y1, double x2, double y2);
int __XpgpX_fLineR(double dx, double dy);
int __XpgpX_fLineTo(double x, double y);

int __XpgpX_Triangle(double x0, double y0, double x1, double y1,
		     double x2, double y2);
int __XpgpX_FilledTriangle(double x0, double y0, double x1, double y1,
		     double x2, double y2);
int __XpgpX_fTriangle(double x0, double y0, double x1, double y1,
		     double x2, double y2);
int __XpgpX_fFilledTriangle(double x0, double y0, double x1, double y1,
		     double x2, double y2);

int __XpgpX_Rectangle(double x, double y, double width, double height);
int __XpgpX_RectangleS(double x1, double y1, double x2, double y2);
int __XpgpX_FilledRectangle(double x, double y, double width, double height);
int __XpgpX_FilledRectangleS(double x1, double y1, double x2, double y2);

int __XpgpX_fRectangle(double x, double y, double width, double height);
int __XpgpX_fRectangleS(double x1, double y1, double x2, double y2);
int __XpgpX_fFilledRectangle(double x, double y, double width, double height);
int __XpgpX_fFilledRectangleS(double x1, double y1, double x2, double y2);

int __XpgpX_Arc(double x, double y,
		double r1, double r2, double a1, double a2);
int __XpgpX_Circle(double x, double y, double r);
int __XpgpX_Ellipse(double x, double y, double r1, double r2);
int __XpgpX_FilledArc(double x, double y,
		      double r1, double r2, double a1, double a2);
int __XpgpX_FilledCircle(double x, double y, double r);
int __XpgpX_FilledEllipse(double x, double y, double r1, double r2);
int __XpgpX_fArc(double x, double y,
		 double r1, double r2, double a1, double a2);
int __XpgpX_fCircle(double x, double y, double r);
int __XpgpX_fEllipse(double x, double y, double r1, double r2);
int __XpgpX_fFilledArc(double x, double y,
		       double r1, double r2, double a1, double a2);
int __XpgpX_fFilledCircle(double x, double y, double r);
int __XpgpX_fFilledEllipse(double x, double y, double r1, double r2);


int __XpgpX_DrawString(double x, double y, char *text);
int __XpgpX_fDrawString(double x, double y, char *text);

int __XpgpX_PutImage(double x, double y, unsigned char *data,
                      unsigned int width,unsigned int height);

int __XpgpX_PutBitmap(double x, double y, char *bitmap,
                      unsigned int width,unsigned int height,int fg,int bg);

int __XpgpX_PlotQP(double x, double y);
int __XpgpX_fPlotQP(double x, double y);

int __XpgpX_AllocateBasicColors(void);



int __XpgpX_CloseX(union __Xpgp_OutputDevice *dev);

/***************************************************************************/

int __XpgpPS_InitPS(char *file_name);

int __XpgpPS_Clear(void);
int __XpgpPS_fShow(void);

int __XpgpPS_GetMouseStatus(int *x, int *y, unsigned int *status);
int __XpgpPS_GetMousePressEvent(int *x, int *y, unsigned int *status);
int __XpgpPS_SetFocus(union __Xpgp_OutputDevice *dev);

int __XpgpPS_Ink(unsigned short red,unsigned short green,unsigned short blue);
int __XpgpPS_ClearPalette(void);
int __XpgpPS_ResetPalette(void);

int __XpgpPS_SetBackground(int color);
int __XpgpPS_Pen(int color);
int __XpgpPS_Paper(int color);

int __XpgpPS_SetLineWidth(double w);
int __XpgpPS_SetDash(double b1, double w1, double b2, double w2);
int __XpgpPS_SetFontSize(double size);
int __XpgpPS_SetQP(double d, int shape);

int __XpgpPS_Move(double x, double y);
int __XpgpPS_MoveR(double dx, double dy);

int __XpgpPS_Plot(double x, double y);
int __XpgpPS_PlotR(double dx, double dy);

int __XpgpPS_Line(double x1, double y1, double x2, double y2);
int __XpgpPS_LineTo(double x, double y);
int __XpgpPS_LineR(double dx, double dy);

int __XpgpPS_Triangle(double x1, double y1, double x2, double y2,
		     double x3, double y3);
int __XpgpPS_FilledTriangle(double x1, double y1, double x2, double y2,
		     double x3, double y3);

int __XpgpPS_Rectangle(double x, double y, double width, double height);
int __XpgpPS_RectangleS(double x1, double y1, double x2, double y2);
int __XpgpPS_FilledRectangle(double x, double y, double width, double height);
int __XpgpPS_FilledRectangleS(double x1, double y1, double x2, double y2);

int __XpgpPS_Arc(double x, double y,
		 double r1, double r2, double a1, double a2);
int __XpgpPS_Circle(double x, double y, double r);
int __XpgpPS_Ellipse(double x, double y, double r1, double r2);
int __XpgpPS_FilledArc(double x, double y,
		       double r1, double r2, double a1, double a2);
int __XpgpPS_FilledCircle(double x, double y, double r);
int __XpgpPS_FilledEllipse(double x, double y, double r1, double r2);

int __XpgpPS_PutImage(double x, double y, unsigned char *data,
                      unsigned int width,unsigned int height);

int __XpgpPS_PutBitmap(double x, double y, char *bitmap,
                      unsigned int width,unsigned int height,int fg,int bg);

int __XpgpPS_DrawString(double x, double y, char *text);

int __XpgpPS_PlotQP(double x, double y);

int __XpgpPS_AllocateBasicColors(void);

int __XpgpPS_ClosePS(union __Xpgp_OutputDevice *dev);


/***************************************************************************/

int __XpgpBMP_InitBMP(char *file_name, int width, int height);

int __XpgpBMP_Clear(void);
int __XpgpBMP_fShow(void);

int __XpgpBMP_GetMouseStatus(int *x, int *y, unsigned int *status);
int __XpgpBMP_GetMousePressEvent(int *x, int *y, unsigned int *status);
int __XpgpBMP_SetFocus(union __Xpgp_OutputDevice *dev);

int __XpgpBMP_Ink(unsigned short red,unsigned short green,unsigned short blue);
int __XpgpBMP_ClearPalette(void);
int __XpgpBMP_ResetPalette(void);

int __XpgpBMP_SetBackground(int color);
int __XpgpBMP_Pen(int color);
int __XpgpBMP_Paper(int color);

int __XpgpBMP_SetLineWidth(double w);
int __XpgpBMP_SetDash(double b1, double w1, double b2, double w2);
int __XpgpBMP_SetFontSize(double size);
int __XpgpBMP_SetQP(double d, int shape);

int __XpgpBMP_Move(double x, double y);
int __XpgpBMP_MoveR(double dx, double dy);

int __XpgpBMP_Plot(double x, double y);
int __XpgpBMP_PlotR(double dx, double dy);

int __XpgpBMP_Line(double x1, double y1, double x2, double y2);
int __XpgpBMP_LineTo(double x, double y);
int __XpgpBMP_LineR(double dx, double dy);

int __XpgpBMP_Triangle(double x1, double y1, double x2, double y2,
		     double x3, double y3);
int __XpgpBMP_FilledTriangle(double x1, double y1, double x2, double y2,
		     double x3, double y3);

int __XpgpBMP_Rectangle(double x, double y, double width, double height);
int __XpgpBMP_RectangleS(double x1, double y1, double x2, double y2);
int __XpgpBMP_FilledRectangle(double x, double y, double width, double height);
int __XpgpBMP_FilledRectangleS(double x1, double y1, double x2, double y2);

int __XpgpBMP_Arc(double x, double y,
		 double r1, double r2, double a1, double a2);
int __XpgpBMP_Circle(double x, double y, double r);
int __XpgpBMP_Ellipse(double x, double y, double r1, double r2);
int __XpgpBMP_FilledArc(double x, double y,
		       double r1, double r2, double a1, double a2);
int __XpgpBMP_FilledCircle(double x, double y, double r);
int __XpgpBMP_FilledEllipse(double x, double y, double r1, double r2);

int __XpgpBMP_PutImage(double x, double y, unsigned char *data,
                      unsigned int width,unsigned int height);

int __XpgpBMP_PutBitmap(double x, double y, char *bitmap,
                      unsigned int width,unsigned int height,int fg,int bg);

int __XpgpBMP_DrawString(double x, double y, char *text);

int __XpgpBMP_PlotQP(double x, double y);

int __XpgpBMP_AllocateBasicColors(void);

int __XpgpBMP_CloseBMP(union __Xpgp_OutputDevice *dev);
