#include "XpgpP.h"

#include "./fonts/Xpgp_BMP_fonts.ac"

#ifndef PI
#define PI 3.14159
#endif

#define BI_RGB  0
#define BI_RLE8 1
#define BI_RLE4 2

#define MIN(a,b) (((a)<(b))? (a):(b))
#define MAX(a,b) (((a)>(b))? (a):(b))


static void __XpgpBMP_PlotP(int x, int y);
static int  __XpgpBMP_GetPixel(int x, int y);
static void __XpgpBMP_PlotPT(int x, int y, int t);
void __XpgpBMP_Fill(int xc, int yc,
		    int x1, int y1, int x2, int y2,
		    int pen_border);
static void __XpgpBMP_FillRecursiv(int x, int y);
int __XpgpBMP_PutBitmapTransparent(double x, double y,char *bitmap,
				   unsigned int width,unsigned int height,
				   int fg);


static void __XpgpBMP_WriteBMPFile();




#define max(a,b) (((a)<(b))? (b):(a))
#define abs(a) (((a)>0)? (a):-(a) )


/*
 *
 *	    Initialize BMP
 *
 */
int __XpgpBMP_InitBMP(char *file_name, int width, int height)
{
  int i;
  if((CurOD->BMP.fp=fopen(file_name,"w"))==NULL) {
    fprintf(stderr,"Xpgp: can not open file  '%s'\n",file_name);
    return 1;
  }
  
  CurOD->BMP.w=width;
  CurOD->BMP.h=height;
  CurOD->BMP.bmp=malloc(width*height*sizeof(int));
  if(CurOD->BMP.bmp==NULL) {
    fprintf(stderr, "Xpgp: BMP: Can not allocate memory\n");
    exit(2);
  }
  for(i=0;i<width*height;i++)		/* clear bitmap */
    CurOD->BMP.bmp[i]=0;
  
  __XpgpBMP_AllocateBasicColors();
  __XpgpBMP_SetDash(0.0,0.0,0.0,0.0);
  __XpgpBMP_SetLineWidth(0);
  __XpgpBMP_Pen(1);
  __XpgpBMP_SetFontSize(12);
  
  return 0;
}


/*
 *
 *    Close BMP output
 *
 */
int __XpgpBMP_CloseBMP(union __Xpgp_OutputDevice *dev)
{
 __XpgpBMP_WriteBMPFile();		/* write BMP file */
 free(CurOD->BMP.bmp);			/* free memory    */
 free(CurOD->BMP.inks);
 return 0;
}



int __XpgpBMP_Clear(void)
{
  int i;
  for(i=0;i<CurOD->BMP.w*CurOD->BMP.h;i++)	  /* clear bitmap */
    CurOD->BMP.bmp[i]=0;
  return 0;
}



int __XpgpBMP_Ink(unsigned short red,unsigned short green,unsigned short blue)
{
 CurOD->BMP.NofInks++;
 if(CurOD->BMP.inks==NULL)
   CurOD->BMP.inks=malloc(CurOD->BMP.NofInks*
			  sizeof(CurOD->BMP.inks[0]));
 else
   CurOD->BMP.inks=realloc((void *)CurOD->BMP.inks,
			   CurOD->BMP.NofInks*
			   sizeof(CurOD->BMP.inks[0]));
 if(CurOD->BMP.inks==NULL) {
   fputs("Xpgp: not enough memory\n",stderr);
   return -1;
 }
 

 CurOD->BMP.inks[CurOD->BMP.NofInks-1].r=red;
 CurOD->BMP.inks[CurOD->BMP.NofInks-1].g=green;
 CurOD->BMP.inks[CurOD->BMP.NofInks-1].b=blue;
 
 return CurOD->BMP.NofInks-1;
}


int __XpgpBMP_ClearPalette(void)
{
 CurOD->BMP.NofInks=0;
 free((void *)CurOD->BMP.inks);
 CurOD->BMP.inks=NULL;
 return 0;
}


int __XpgpBMP_ResetPalette(void)
{
 __XpgpBMP_ClearPalette();
 __XpgpBMP_AllocateBasicColors();
 __XpgpBMP_Pen(1);
 return 0;
}



int __XpgpBMP_Pen(int color)
{
 if(color>=CurOD->BMP.NofInks || color<0)
     return -1;
 CurOD->BMP.curPen=color;
 return 0;
}



/*
 *
 *	    Set Line Width
 *
 */
int __XpgpBMP_SetLineWidth(double w)
{
 CurOD->BMP.lw=w;
 return 0;
}


int __XpgpBMP_SetDash(double b1,double w1,double b2,double w2)
{
 return 0;
}



int __XpgpBMP_SetFontSize(double size)
{
 return 0;
}


int __XpgpBMP_Move(double x, double y)
{
    CurOD->BMP.gcur.x=x;
    CurOD->BMP.gcur.y=y;
    return 0;
}


int __XpgpBMP_MoveR(double dx, double dy)
{
    CurOD->BMP.gcur.x+=dx;
    CurOD->BMP.gcur.y+=dy;
    return 0;
}


     
/* 
 *  Private plot
 */ 
void __XpgpBMP_PlotP(int x, int y)
{
    if(x>=0 && x<CurOD->BMP.w && y>=0 && y<CurOD->BMP.h)
	CurOD->BMP.bmp[x+y*CurOD->BMP.w]=CurOD->BMP.curPen;
}

/*
 *  Get pixel
 */
int __XpgpBMP_GetPixel(int x, int y)
{
    if(x>=0 && x<CurOD->BMP.w && y>=0 && y<CurOD->BMP.h)
	return CurOD->BMP.bmp[x+y*CurOD->BMP.w];
    else
	return -1;
}


/* 
 *  Private plot with thickness
 */ 
void __XpgpBMP_PlotPT(int x, int y, int t)
{
    int th[3][32]={
	{1, 0,0},
	{4, 0,0, 1,0, 0,1, 1,1},
	{9, 0,0, 1,0, 1,-1, 0,-1, -1,-1, -1,0, -1,1, 0,1, 1,1}
    };
    int i;

    t--;
    if(t<0) t=0;
    if(t>2) t=2;

    for(i=0;i<th[t][0];i++)
	__XpgpBMP_PlotP(x+th[t][i*2+1], y+th[t][i*2+2]);
    
}


     
int __XpgpBMP_Plot(double x, double y)
{
    CurOD->BMP.gcur.x=x;
    CurOD->BMP.gcur.y=y;
    
    __XpgpBMP_PlotP(dtoi(x), dtoi(y));		 
    return 0;
}


int __XpgpBMP_PlotR(double dx, double dy)
{
    CurOD->BMP.gcur.x+=dx;
    CurOD->BMP.gcur.y+=dy;
    __XpgpBMP_PlotP(dtoi(CurOD->BMP.gcur.x), dtoi(CurOD->BMP.gcur.y));
    return 0;
}


int __XpgpBMP_Line(double x1, double y1, double x2, double y2)
{
    int i, d, dx, dy, x, y;
    CurOD->BMP.gcur.x=x2;
    CurOD->BMP.gcur.y=y2;
    dx=x2-x1;
    dy=y2-y1;
    d=max(abs(dx), abs(dy));
    if(d==0)
	d=1;
    
    for(i=0;i<=d;i++) {
	x=x1+(dx*i)/d;
	y=y1+(dy*i)/d;
	__XpgpBMP_PlotPT(x,y,CurOD->BMP.lw);
    }
    
    return 0;
}


int __XpgpBMP_LineTo(double x, double y)
{
    __XpgpBMP_Line(CurOD->BMP.gcur.x, CurOD->BMP.gcur.y, x, y);
    return 0;
}


int __XpgpBMP_LineR(double dx, double dy)
{
    __XpgpBMP_Line(CurOD->BMP.gcur.x, CurOD->BMP.gcur.y,
		   CurOD->BMP.gcur.x+dx, CurOD->BMP.gcur.y+dy);
    return 0;
}


/*
 *  Draw Triangle
 */
int __XpgpBMP_Triangle(double x1, double y1,
		       double x2, double y2,
		       double x3, double y3)
{
    __XpgpBMP_Line(x1,y1, x2,y2);
    __XpgpBMP_Line(x2,y2, x3,y3);
    __XpgpBMP_Line(x3,y3, x1,y1);
    return 0;
}


int __XpgpBMP_FilledTriangle(double x1, double y1,
			     double x2, double y2,
			     double x3, double y3)
{
    int pen_sto;
    int yl, xl, xh, yh;

    pen_sto=CurOD->BMP.curPen;
    CurOD->BMP.curPen=-1;
    __XpgpBMP_Line(x1,y1, x2,y2);
    __XpgpBMP_Line(x2,y2, x3,y3);
    __XpgpBMP_Line(x3,y3, x1,y1);
    CurOD->BMP.curPen=pen_sto;

    xl=MIN(MIN(x1, x2), x3);
    yl=MIN(MIN(y1, y2), y3);
    xh=MAX(MAX(x1, x2), x3);
    yh=MAX(MAX(y1, y2), y3);
    
    __XpgpBMP_Fill((x1+x2+x3)/3, (y1+y2+y3)/3,	  /* slow, but works */
	           xl, yl, xh, yh,
	           -1);
    return 0;
}


int __XpgpBMP_Rectangle(double x, double y, double width, double height)
{
    __XpgpBMP_RectangleS(x,y,x+width,y+height);
    return 0;
}


int __XpgpBMP_RectangleS(double x1, double y1, double x2, double y2)
{
    __XpgpBMP_Line(x1,y1, x2,y1);
    __XpgpBMP_Line(x2,y1, x2,y2);
    __XpgpBMP_Line(x2,y2, x1,y2);
    __XpgpBMP_Line(x1,y2, x1,y1);
    return 0;
}


int __XpgpBMP_FilledRectangle(double x, double y, double width, double height)
{
    int i, ixw, ix, iy, iw, ih;

    ix=x;
    iy=y;
    iw=width;
    ih=height;
    ixw=ix+iw;

    if(ih>0)
	for(i=0;i<=ih;i++)
	    __XpgpBMP_Line(ix,i+iy, ixw, i+iy);
    else
	for(i=ih;i>=0;i--)
	    __XpgpBMP_Line(ix,i+iy, ixw, i+iy);

    return 0;
}


int __XpgpBMP_FilledRectangleS(double x1, double y1, double x2, double y2)
{
    __XpgpBMP_FilledRectangle(x1,y1,x2-x1,y2-y1);
    return 0;
}


int __XpgpBMP_Arc(double x, double y,
		 double r1, double r2, double a1, double a2)
{
    long a, b, g, dgp, dgm;
    long xi,yi,x0;
    double an;

    if(r1==0) {
	__XpgpBMP_Line(x,y-r2, x,y+r2);
	return 0;
    }
    if(r2==0) {
	__XpgpBMP_Line(x-r1,y, x+r1,y);
	return 0;
    }
    
    a1=fmod(a1,2.*360.);
    a2=fmod(a2,2.*360.);
    if(a1>a2) {
	an=a1;
	a1=a2;
	a2=an;
    }
    
    a1=a1*PI/180.0;
    a2=a2*PI/180.0;


    a=r1;
    b=r2;
    x0=(double)a*a/sqrt(a*a+b*b);    
    xi=0;
    yi=b;
    g = b*b + a*a * (yi*yi-yi) - a*a*b*b;
    dgp = 3*b*b - 2*a*a*(b-1);
    dgm = 3*b*b;

    while(xi<=x0) {
	if(xi!=0.0)
	    an=atan((double)yi/xi);
	else
	    an=PI/2;
	if(2.0*PI-an>a1 && 2.0*PI-an<a2)	
	    __XpgpBMP_PlotPT(x+xi,y+yi,CurOD->BMP.lw);
	if(PI+an>a1 && PI+an<a2)
	    __XpgpBMP_PlotPT(x-xi,y+yi,CurOD->BMP.lw);
	if(PI-an>a1 && PI-an<a2)
	    __XpgpBMP_PlotPT(x-xi,y-yi,CurOD->BMP.lw);
	if(an>a1 && an<a2)
	    __XpgpBMP_PlotPT(x+xi,y-yi,CurOD->BMP.lw);

	if(g>0) {
	    g+=dgp;
	    dgp+=2*b*b+2*a*a;
	    dgm+=2*b*b;
	    xi++;
	    yi--;
	} else {
	    g+=dgm;
	    dgp+=2*b*b;
	    dgm+=2*b*b;
	    xi++;
	}
    }


    
    a=r2;
    b=r1;
    xi=0;
    yi=b;
    g = b*b + a*a * (yi*yi-yi) - a*a*b*b;
    dgp = 3*b*b - 2*a*a*(b-1);
    dgm = 3*b*b;

    while(yi>=x0) {
	if(yi!=0.0)
	    an=atan((double)xi/yi);
	else
	    an=PI/2;
	if(2.0*PI-an>a1 && 2.0*PI-an<a2)
	    __XpgpBMP_PlotPT(x+yi,y+xi,CurOD->BMP.lw);
	if(PI+an>a1 && PI+an<a2)
	    __XpgpBMP_PlotPT(x-yi,y+xi,CurOD->BMP.lw);
	if(PI-an>a1 && PI-an<a2)
	    __XpgpBMP_PlotPT(x-yi,y-xi,CurOD->BMP.lw);
	if(an>a1 && an<a2)       
	    __XpgpBMP_PlotPT(x+yi,y-xi,CurOD->BMP.lw);
	if(g>0) {
	    g+=dgp;
	    dgp+=2*b*b+2*a*a;
	    dgm+=2*b*b;
	    xi++;
	    yi--;
	} else {
	    g+=dgm;
	    dgp+=2*b*b;
	    dgm+=2*b*b;
	    xi++;
	}
    }

    
    return 0;
}
 

int __XpgpBMP_Circle(double x, double y, double r)
{
    __XpgpBMP_Ellipse(x,y,r,r);
    return 0;
}
 

int __XpgpBMP_Ellipse(double x, double y, double r1, double r2)
{
    long a, b, g, dgp, dgm;
    long xi,yi,x0,y0;

    if(r1==0) {
	__XpgpBMP_Line(x,y-r2, x,y+r2);
	return 0;
    }
    if(r2==0) {
	__XpgpBMP_Line(x-r1,y, x+r1,y);
	return 0;
    }
    
    a=r1;
    b=r2;
    x0=(long)((double)a*a/sqrt((double)a*a+b*b));    
    xi=0;
    yi=b;
    g = b*b + a*a * (yi*yi-yi) - a*a*b*b;
    dgp = 3*b*b - 2*a*a*(b-1);
    dgm = 3*b*b;

    while(xi<=x0) {
	__XpgpBMP_PlotPT(x+xi,y+yi,CurOD->BMP.lw);
	__XpgpBMP_PlotPT(x-xi,y+yi,CurOD->BMP.lw);
	__XpgpBMP_PlotPT(x+xi,y-yi,CurOD->BMP.lw);
	__XpgpBMP_PlotPT(x-xi,y-yi,CurOD->BMP.lw);
	if(g>0) {
	    g+=dgp;
	    dgp+=2*b*b+2*a*a;
	    dgm+=2*b*b;
	    xi++;
	    yi--;
	} else {
	    g+=dgm;
	    dgp+=2*b*b;
	    dgm+=2*b*b;
	    xi++;
	}
    }

    a=r2;
    b=r1;
    y0=(double)b*b/sqrt(a*a+b*b);    
    xi=0;
    yi=b;
    g = b*b + a*a * (yi*yi-yi) - a*a*b*b;
    dgp = 3*b*b - 2*a*a*(b-1);
    dgm = 3*b*b;

    while(yi>=x0) {
	__XpgpBMP_PlotPT(x+yi,y+xi,CurOD->BMP.lw);
	__XpgpBMP_PlotPT(x-yi,y+xi,CurOD->BMP.lw);
	__XpgpBMP_PlotPT(x+yi,y-xi,CurOD->BMP.lw);
	__XpgpBMP_PlotPT(x-yi,y-xi,CurOD->BMP.lw);
	if(g>0) {
	    g+=dgp;
	    dgp+=2*b*b+2*a*a;
	    dgm+=2*b*b;
	    xi++;
	    yi--;
	} else {
	    g+=dgm;
	    dgp+=2*b*b;
	    dgm+=2*b*b;
	    xi++;
	}
    }
    
    return 0;
}



int __XpgpBMP_FilledArc(double x, double y,
		       double r1, double r2, double a1, double a2)
{ 
    double t1, t2; 
    int old_pen;

    old_pen=CurOD->BMP.curPen;

    CurOD->BMP.curPen=-1;
    
    __XpgpBMP_Arc(x,y,r1,r2,a1,a2);

    r1+=0.5;
    r2+=0.5;

    a1=fmod(a1*PI/180.0, 2*PI); 
    a2=fmod(a2*PI/180.0, 2*PI);
    
    t1=atan(r1/r2 * tan(a1));
    t2=atan(r1/r2 * tan(a2));


    if(a1<PI/2)					  /* atan=[-PI/2, PI/2] */
	t1=t1;
    else if(a1<PI)
	t1=PI+t1;
    else if(a1<3*PI/2)
	t1=PI+t1;
    else
	t1=2*PI+t1;
    
    if(a2<PI/2)
	t2=t2;
    else if(a2<PI)
	t2=PI+t2;
    else if(a2<3*PI/2)
	t2=PI+t2;
    else
	t2=2*PI+t2;
    
    __XpgpBMP_Line(x,y, x+r1*cos(t1), y-r2*sin(t1));
    __XpgpBMP_Line(x,y, x+r1*cos(t2), y-r2*sin(t2));

    CurOD->BMP.curPen=old_pen;			  /* reset pen */
    
    __XpgpBMP_Fill(x+0.5*r1*cos((t1+t2)/2),	  /* recursive fill */
		   y-0.5*r2*sin((t1+t2)/2),
		   x-r1, y-r2, x+r1, y+r2,
		   -1);
    
    return 0; 
} 


 

int __XpgpBMP_FilledCircle(double x, double y, double r)
{
    __XpgpBMP_FilledEllipse(x,y,r,r);
    return 0;
}
 

int __XpgpBMP_FilledEllipse(double x, double y, double r1, double r2)
{
    long a, b, g, dgp, dgm;
    long xi,yi,x0, y0;

    if(r1==0) {
	__XpgpBMP_Line(x,y-r2, x,y+r2);
	return 0;
    }
    if(r2==0) {
	__XpgpBMP_Line(x-r1,y, x+r1,y);
	return 0;
    }
    
    a=r1;
    b=r2;
    x0=(double)a*a/sqrt(a*a+b*b);    
    xi=0;
    yi=b;
    g = b*b + a*a * (yi*yi-yi) - a*a*b*b;
    dgp = 3*b*b - 2*a*a*(b-1);
    dgm = 3*b*b;

    while(xi<=x0) {
	__XpgpBMP_Line(x+xi,y+yi, x+xi,y-yi);
	__XpgpBMP_Line(x-xi,y+yi, x-xi,y-yi);
	if(g>0) {
	    g+=dgp;
	    dgp+=2*b*b+2*a*a;
	    dgm+=2*b*b;
	    xi++;
	    yi--;
	} else {
	    g+=dgm;
	    dgp+=2*b*b;
	    dgm+=2*b*b;
	    xi++;
	}
    }

    
    a=r2;
    b=r1;
    y0=(double)b*b/sqrt(a*a+b*b);    
    xi=0;
    yi=b;
    g = b*b + a*a * (yi*yi-yi) - a*a*b*b;
    dgp = 3*b*b - 2*a*a*(b-1);
    dgm = 3*b*b;

    while(yi>=x0) {
	__XpgpBMP_Line(x+yi,y+xi, x+yi,y-xi);
	__XpgpBMP_Line(x-yi,y+xi, x-yi,y-xi);
	if(g>0) {
	    g+=dgp;
	    dgp+=2*b*b+2*a*a;
	    dgm+=2*b*b;
	    xi++;
	    yi--;
	} else {
	    g+=dgm;
	    dgp+=2*b*b;
	    dgm+=2*b*b;
	    xi++;
	}
    }

    
    return 0;
}

  

int __XpgpBMP_DrawString(double x, double y, char *text)
{
    int i, l, c, inc;
    l=strlen(text);
    for(i=0;i<l;i++) {
	c=text[i];
	inc=__XpgpBMP_PutBitmapTransparent(x, y-16, &__XpgpBMP_font[c][0],
					   16, 16,
					   CurOD->BMP.curPen);
	if(inc)
	    x+=inc+2;
	else
	    x+=16/4;
    }
    return 0;
}



int __XpgpBMP_SetQP(double d, int shape)
{
    CurOD->BMP.QPd=d;
    CurOD->BMP.QPshape=shape;
    return 0;
}


int __XpgpBMP_PlotQP(double x, double y)
{
     double d;
     int lw_sto;
     
     lw_sto=CurOD->BMP.lw;
     CurOD->BMP.lw=1;
     
     d=CurOD->BMP.QPd;
     x*=d;
     y*=d;

     
     switch(CurOD->BMP.QPshape) {
       case 0:
	 __XpgpBMP_FilledRectangle(x-d/2,y-d/2,d-1,d-1);
	 break;
       case 1:
	 __XpgpBMP_FilledCircle(x,y,d/2);
	 break;
       default:
	 fprintf(stderr, "Xpgp: unknown QP shape (%d)!\n", CurOD->BMP.QPshape);
	 return -1;
     }

     CurOD->BMP.lw=lw_sto;
     return 0;
}



/*
 *
 * data is pen array
 *
 */
int __XpgpBMP_PutImage(double x, double y, unsigned char *data,
                      unsigned int width,unsigned int height)
{
    int pen_sto, i, j;

    pen_sto=CurOD->BMP.curPen;

    for(j=0;j<height;j++)
	for(i=0;i<width;i++) {
	    if(__XpgpBMP_Pen(data[i+j*width])!=0)
		fprintf(stderr, "Xpgp: PutImage: Pen %d not allocated!\n",
			data[i+j*width]);
	    __XpgpBMP_PlotP(x+i, y+j);
	}
    
    CurOD->BMP.curPen=pen_sto;
    
    return 0;
}



int __XpgpBMP_PutBitmap(double x, double y,char *bitmap,
                      unsigned int width,unsigned int height,
                      int fg,int bg)
{
    int pen_sto, i, j, n, byte, inc;
/*     unsigned char mask_lh[8]={0x01, 0x02, 0x04, 0x08, */
/* 			      0x10, 0x20, 0x40, 0x80};   */
    unsigned char mask_hl[8]={0x80, 0x40, 0x20, 0x10,
			      0x08, 0x04, 0x02, 0x01};

    pen_sto=CurOD->BMP.curPen;
    
    n=0;
    
    if(width%8)
	inc=8-width%8;
    else
	inc=0;
    
    for(i=0;i<height;i++) {
	for(j=0;j<width;j++) {
	    byte = j/8 + (i*(width+inc))/8;
	    if(mask_hl[n%8]&bitmap[byte])
		__XpgpBMP_Pen(fg);
	    else
		__XpgpBMP_Pen(bg);
	    __XpgpBMP_PlotP(j+x,i+y);
	    n++;
	}
	n+=inc;
    }
    
    CurOD->BMP.curPen=pen_sto;
    
    return 0;
}


/*
 * return x coor. of most right bit
 */
int __XpgpBMP_PutBitmapTransparent(double x, double y,char *bitmap,
				   unsigned int width,unsigned int height,
				   int fg)
{
    int pen_sto, i, j, n, byte, inc;
    unsigned char mask_hl[8]={0x80, 0x40, 0x20, 0x10,
			      0x08, 0x04, 0x02, 0x01};
    int maxx;

    pen_sto=CurOD->BMP.curPen;
    
    n=0;
    
    if(width%8)
	inc=8-width%8;
    else
	inc=0;

    maxx=0;
    __XpgpBMP_Pen(fg);
    for(i=0;i<height;i++) {
	for(j=0;j<width;j++) {
	    byte = j/8 + (i*(width+inc))/8;
	    if(mask_hl[n%8]&bitmap[byte]) {
		__XpgpBMP_PlotP(j+x,i+y);
		if(j>maxx)
		    maxx=j;
	    }
	    n++;
	}
	n+=inc;
    }
    
    CurOD->BMP.curPen=pen_sto;
    
    return maxx;
}


/*****************************************************************************

			    GENERAL

 *****************************************************************************/

static int BMP_fill_border;

/*
 * Fill shape defined with pen_border
 */
void __XpgpBMP_Fill(int xc, int yc,
		    int x1, int y1, int x2, int y2,
		    int pen_border)
{
    int i,j;
    int pen_sto;

    if(x1>x2) {
	i=x1; x1=x2; x2=i;
    }
    if(y1>y2) {
	i=y1; y1=y2; y2=i;
    }
    
    pen_sto=CurOD->BMP.curPen;

    CurOD->BMP.curPen=pen_border;
    BMP_fill_border=pen_border;
    
    __XpgpBMP_FillRecursiv(xc,yc);

    CurOD->BMP.curPen=pen_sto;
    
    for(i=x1;i<=x2;i++)
	for(j=y1;j<=y2;j++)
	    if(__XpgpBMP_GetPixel(i,j)==pen_border)
		__XpgpBMP_PlotP(i,j);    
}


void __XpgpBMP_FillRecursiv(int x, int y)
{
    __XpgpBMP_Plot(x,y);
    
    if(__XpgpBMP_GetPixel(x+1,y)==BMP_fill_border)
	__XpgpBMP_Plot(x+1,y);
    else
	__XpgpBMP_FillRecursiv(x+1,y);

    if(__XpgpBMP_GetPixel(x,y-1)==BMP_fill_border)
	__XpgpBMP_Plot(x,y-1);
    else
	__XpgpBMP_FillRecursiv(x,y-1);

    if(__XpgpBMP_GetPixel(x-1,y)==BMP_fill_border)
	__XpgpBMP_Plot(x-1,y);
    else
	__XpgpBMP_FillRecursiv(x-1,y);

    if(__XpgpBMP_GetPixel(x,y+1)==BMP_fill_border)
	__XpgpBMP_Plot(x,y+1);
    else
	__XpgpBMP_FillRecursiv(x,y+1);
    return;
}
    


int __XpgpBMP_AllocateBasicColors(void)
{
/*
 *   "White","Black",
 *   "Light Blue","Cyan","Blue","Navy","Violet",
 *   "Magenta","Purple","Pink","Khaki","Aquamarine",
 *   "Pale Green","Green","Lime Green","Yellow Green","Green Yellow","Yellow",
 *   "Gold","Light Goldenrod","Goldenrod","Dark Goldenrod",
 *   "Brown","Chocolate","Peru","Coral",
 *   "Light Coral","Salmon",
 *   "Dark Orange","Orange","Orange Red",
 *   "Red"
 */
 int i;
 static unsigned short colors[32][3] = {
 {65535,65535,65535},{0,0,0},
 {44461,55512,59110},{0,65535,65535},{0,0,65535},{0,0,32896},{61166,33410,61166},
 {65535,0,65535},{41120,8224,61680},{65535,49344,52171},{61680,59110,35980},{32639,65535,54484},
 {39064,64507,39064},{0,65535,0},{12850,52685,12850},{39578,52685,12850},{44461,65535,12079},{65535,65535,0},
 {65535,55255,0},{61166,56797,33410},{56026,42405,8224},{47288,34438,2827},
 {42405,10794,10794},{53970,26985,7710},{52685,34181,16191},{65535,32639,20560},
 {61680,32896,32896},{64250,32896,29298},
 {65535,35980,0},{65535,42405,0},{65535,17733,0},
 {65535,0,0}};

 if(CurOD->BMP.inks!=NULL)
   free(CurOD->BMP.inks);
 if((CurOD->BMP.inks=malloc(32*sizeof(CurOD->BMP.inks[0])))
    ==NULL) {
   fputs("Xpgp: not enough memory\n",stderr);
   return -1;
 }
 CurOD->BMP.NofInks=32;
 for(i=0;i<32;i++) {
     CurOD->BMP.inks[i].r=colors[i][0];
     CurOD->BMP.inks[i].g=colors[i][1];
     CurOD->BMP.inks[i].b=colors[i][2];
 }
 return 0;
}


/*****************************************************************************/


static void put_INT2(FILE *fp, int i)
{
  int cl, ch;
  cl = (((unsigned int) i)   ) & 0xff;
  ch = (((unsigned int) i)>>8) & 0xff;
  putc(cl,fp);
  putc(ch,fp);
}


static void put_INT4(FILE *fp, int i)
{
  int c0, c1, c2, c3;
  c0 = (((unsigned int) i)    ) & 0xff;  
  c1 = (((unsigned int) i)>>8 ) & 0xff;
  c2 = (((unsigned int) i)>>16) & 0xff;
  c3 = (((unsigned int) i)>>24) & 0xff;
  putc(c0,fp);
  putc(c1,fp);
  putc(c2,fp);
  putc(c3,fp);
}


void __XpgpBMP_WriteBMPFile()		/* write BMP file */
{
  int i, j, w, h, n_pad_bytes;
  int file_size, nbits, Nink, byte_lin;
  FILE *fp;
  
  fp=CurOD->BMP.fp;
  Nink=CurOD->BMP.NofInks;
  w=CurOD->BMP.w;
  h=CurOD->BMP.h;

  nbits = 24;
  
  byte_lin=3*w;				/* bytes in line */
  
  file_size = 14 +			/* size of bitmap file header */
              40 +			/* size of bitmap info header */
              byte_lin*h;		/* size of image data */

					/* write BMP file */


					/* BITMAPFILEHEADER  (14 bytes) */
  putc('B', fp);			/* bfType */
  putc('M', fp);
  put_INT4(fp, file_size);		/* bfSize (bytes) */
  put_INT2(fp, 0);			/* bfReserved1 */
  put_INT2(fp, 0);			/* bfReserved2 */
  put_INT4(fp, 14 + 40 );	        /* bfOffBits */

					/* BITMAPINFOHEADER (40 bytes)*/
  put_INT4(fp, 40);			/* biSize */
  put_INT4(fp, w);			/* biWidth */
  put_INT4(fp, h);			/* biHeight */
  put_INT2(fp, 1);			/* biPlanes */
  put_INT2(fp, nbits);			/* biBitCount */
  put_INT4(fp, BI_RGB);			/* biCompression */
  put_INT4(fp, byte_lin*h);		/* biSizeImage */
  put_INT4(fp, 75 * 39);		/* biXPelsPerMeter */
  put_INT4(fp, 75 * 39);		/* biYPelsPerMeter */
  put_INT4(fp, 0);			/* biClrUsed */
  put_INT4(fp, Nink);			/* biClrImportant */
    

					/* RGBQUAD */
  /* skiped (24 BMP) */

					/* 24BMP => (0x00ffffff) */
  n_pad_bytes=(4-(3*w)%4)%4;		/* pad to 4 byte marigin */
  for(j=h-1;j>=0;j--) {
    for(i=0;i<w;i++) {
      putc(CurOD->BMP.inks[ CurOD->BMP.bmp[i+j*w] ].b, fp);
      putc(CurOD->BMP.inks[ CurOD->BMP.bmp[i+j*w] ].g, fp);
      putc(CurOD->BMP.inks[ CurOD->BMP.bmp[i+j*w] ].r, fp);
    }
    for(i=0;i<n_pad_bytes;i++)		/* pad to 32 bit (4byte) */
      putc(0, fp);
  }
}


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

int __XpgpBMP_GetMouseStatus(int *x, int *y, unsigned int *status)
{
 return 0;
}

int __XpgpBMP_GetMousePressEvent(int *x, int *y, unsigned int *status)
{
 return 0;
}

int __XpgpBMP_fShow()
{
 return 0;
}

int __XpgpBMP_SetBackground(int color)
{
 return 0;
}


int __XpgpBMP_SetFocus(union __Xpgp_OutputDevice *dev)
{
  return 0;
}
