/*****************************************************************************
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/
/*                                                                           */
/*                                   #                                       */
/*                                   #                                       */
/*                                  ###                                      */
/*                                  ###                                      */
/*                                 #####                                     */
/*                                 #####                                     */
/*                                #######                                    */
/*                                #######                                    */
/*                   #################################                       */
/*                       #########################                           */
/*                           #################                               */
/*                  [[[[The Working Class Unlimited]]]                       */
/*                             #############                                 */
/*                             #############                                 */
/*                            ######   ######                                */
/*                            ###         ####                               */
/*                           ##              ##                              */
/*                          #                  #                             */
/*                                                                           */
/*                 (C) The Working Class Unlimited 1993.                     */

/**
 **
 **
 **                X p g p  -  P r i m i t i v e s   v2.27
 **
 **
 **                         PostScript - P a r t
 **
 **                                                  Milanovic Ljubomir, 1995.
 **/
#include "XpgpP.h"

#define XpgpPS_plot             (1ul<<0)	/* flags */
#define XpgpPS_triangle         (1ul<<1)
#define XpgpPS_filledTriangle   (1ul<<2)
#define XpgpPS_rectangle        (1ul<<3)
#define XpgpPS_filledRectangle  (1ul<<4)
#define XpgpPS_arc              (1ul<<5)
#define XpgpPS_filledArc        (1ul<<6)
#define XpgpPS_drawString       (1ul<<7)


				/* Initialization functions */
int __XpgpPS_iPlot();
int __XpgpPS_iTriangle();
int __XpgpPS_iFilledTriangle();
int __XpgpPS_iRectangle();
int __XpgpPS_iFilledRectangle();
int __XpgpPS_iArc();
int __XpgpPS_iFilledArc();
int __XpgpPS_iDrawString();


/*
 *
 *	    Initialize PS
 *
 */
int __XpgpPS_InitPS(char *file_name)
{
 if((CurOD->PS.fp=fopen(file_name,"w"))==NULL) {
   fprintf(stderr,"Xpgp: can not open file  '%s'\n",file_name);
   return 1;
 }
 fprintf(CurOD->PS.fp,"%%!PS-Adobe-2.0\n");
 fprintf(CurOD->PS.fp,"%%%%\n");
 fprintf(CurOD->PS.fp,"%%%% Creator: Xpgp\n");
 fprintf(CurOD->PS.fp,"%%%%\n");
 fprintf(CurOD->PS.fp,"newpath\n");

 switch(CurOD->PS.orient) {
   case 0:					/* portrait */
     fprintf(CurOD->PS.fp,"-90 rotate -841 0 translate\n");
     break;
   case '1':					/* landscape */
     break;
   default:
     break;
 }
 __XpgpPS_AllocateBasicColors();
 __XpgpPS_SetDash(0.0,0.0,0.0,0.0);
 __XpgpPS_SetLineWidth(0);
 __XpgpPS_Pen(1);
 __XpgpPS_SetFontSize(12);

 return 0;
}


/*
 *
 *    Close PS output
 *
 */
int __XpgpPS_ClosePS(union __Xpgp_OutputDevice *dev)
{
 fprintf(dev->PS.fp,"\nshowpage\n");
 fprintf(CurOD->PS.fp,"%%%%Page: %d %d\n",
	 CurOD->PS.pageCounter,CurOD->PS.pageCounter);
 fclose(dev->PS.fp);
 free(dev->PS.inks);
 return 0;
}



int __XpgpPS_Clear(void)
{
 fprintf(CurOD->PS.fp,"gsave showpage\n%%%%Page: %d %d\ngrestore newpath\n",
	 CurOD->PS.pageCounter,CurOD->PS.pageCounter);
 CurOD->PS.pageCounter++;
 return 0;
}



int __XpgpPS_Ink(unsigned short red,unsigned short green,unsigned short blue)
{
 CurOD->PS.NofInks++;
 if(CurOD->PS.inks==NULL)
   CurOD->PS.inks=(struct inks *)malloc(CurOD->PS.NofInks*
					sizeof(CurOD->PS.inks[0]));
 else
   CurOD->PS.inks=(struct inks *)realloc((void *)CurOD->PS.inks,
					 CurOD->PS.NofInks*
					 sizeof(CurOD->PS.inks[0]));
 if(CurOD->PS.inks==NULL) {
   fputs("Xpgp: not enough memory\n",stderr);
   return -1;
 }
 

 CurOD->PS.inks[CurOD->PS.NofInks-1].r=red;
 CurOD->PS.inks[CurOD->PS.NofInks-1].g=green;
 CurOD->PS.inks[CurOD->PS.NofInks-1].b=blue;
 
 return CurOD->PS.NofInks-1;
}


int __XpgpPS_ClearPalette(void)
{
 CurOD->PS.NofInks=0;
 free((void *)CurOD->PS.inks);
 CurOD->PS.inks=NULL;
 return 0;
}


int __XpgpPS_ResetPalette(void)
{
 __XpgpPS_ClearPalette();
 __XpgpPS_AllocateBasicColors();
 __XpgpPS_Pen(1);
 return 0;
}



int __XpgpPS_Pen(int color)
{
 if(color>=CurOD->PS.NofInks || color<0)
     return -1;
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f setrgbcolor\n",
	 (double)CurOD->PS.inks[color].r/65535.0,
	 (double)CurOD->PS.inks[color].g/65535.0,
	 (double)CurOD->PS.inks[color].b/65535.0);
 CurOD->PS.curPen=color;
 return 0;
}



/*
 *
 *	    Set Line Width
 *
 */
int __XpgpPS_SetLineWidth(double w)
{
 fprintf(CurOD->PS.fp,"%.3f setlinewidth\n",w);
 return 0;
}


int __XpgpPS_SetDash(double b1,double w1,double b2,double w2)
{
 if(b1>0.0 && w1>0.0 && b2>0.0 && w2>0.0)
     fprintf(CurOD->PS.fp,"[%.3f %.3f %.3f %.3f] 0 setdash\n",
	     b1,w1,b2,w2);
 else
     fprintf(CurOD->PS.fp,"[] 0 setdash\n");
 return 0;
}



int __XpgpPS_SetFontSize(double size)
{
 fprintf(CurOD->PS.fp,"%s findfont %.3f scalefont setfont\n",
	 XPGP_PSFONT,size);
 return 0;
}


int __XpgpPS_Move(double x, double y)
{
 CurOD->PS.gcur.x=x;
 CurOD->PS.gcur.y=y;
 return 0;
}


int __XpgpPS_MoveR(double dx, double dy)
{
 CurOD->PS.gcur.x+=dx;
 CurOD->PS.gcur.y+=dy;
 return 0;
}


int __XpgpPS_Plot(double x, double y)
{
 CurOD->PS.gcur.x=x;
 CurOD->PS.gcur.y=y;
 if(!(CurOD->PS.sr_flag & XpgpPS_plot)) {
     CurOD->PS.sr_flag|=XpgpPS_plot;
     __XpgpPS_iPlot();
 }
 fprintf(CurOD->PS.fp,"%.3f %.3f PL\n",y,x);
 return 0;
}


int __XpgpPS_PlotR(double dx, double dy)
{
 CurOD->PS.gcur.x+=dx;
 CurOD->PS.gcur.y+=dy;
 fprintf(CurOD->PS.fp,"%.3f %.3f PL\n",CurOD->PS.gcur.x,CurOD->PS.gcur.y);
 return 0;
}


int __XpgpPS_Line(double x1, double y1, double x2, double y2)
{
 CurOD->PS.gcur.x=x2;
 CurOD->PS.gcur.y=y2;
 fprintf(CurOD->PS.fp,"%.3f %.3f moveto %.3f %.3f lineto stroke\n",y1,x1,y2,x2);
 return 0;
}


int __XpgpPS_LineTo(double x, double y)
{
 fprintf(CurOD->PS.fp,"%.3f %.3f moveto %.3f %.3f lineto stroke\n",
	 CurOD->PS.gcur.y,CurOD->PS.gcur.x,y,x);
 CurOD->PS.gcur.x=x;
 CurOD->PS.gcur.y=y;
 return 0;
}


int __XpgpPS_LineR(double dx, double dy)
{
 fprintf(CurOD->PS.fp,"%.3f %.3f moveto %.3f %.3f rlineto stroke\n",
	  CurOD->PS.gcur.y,CurOD->PS.gcur.x,dy,dx);
 CurOD->PS.gcur.x+=dx;
 CurOD->PS.gcur.y+=dy;
 return 0;
}


/*
 *  Draw Triangle
 */
int __XpgpPS_Triangle(double x1, double y1,
		      double x2, double y2,
		      double x3, double y3)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_triangle)) {
     CurOD->PS.sr_flag|=XpgpPS_triangle;
     __XpgpPS_iTriangle();
 }
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f %.3f %.3f %.3f TR\n",
	 y1,x1,y2,x2,y3,x3);
 return 0;
}


int __XpgpPS_FilledTriangle(double x1, double y1,
		      double x2, double y2,
		      double x3, double y3)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_filledTriangle)) {
     CurOD->PS.sr_flag|=XpgpPS_filledTriangle;
     __XpgpPS_iFilledTriangle();
 }
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f %.3f %.3f %.3f FTR\n",
	 y1,x1,y2,x2,y3,x3);
 return 0;
}


int __XpgpPS_Rectangle(double x, double y, double width, double height)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_rectangle)) {
     CurOD->PS.sr_flag|=XpgpPS_rectangle;
     __XpgpPS_iRectangle();
 }
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f %.3f RC\n",
	 height, width, y, x);
 return 0;
}


int __XpgpPS_RectangleS(double x1, double y1, double x2, double y2)
{
 __XpgpPS_Rectangle(x1,y1,x2-x1,y2-y1);
 return 0;
}


int __XpgpPS_FilledRectangle(double x, double y, double width, double height)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_filledRectangle)) {
     CurOD->PS.sr_flag|=XpgpPS_filledRectangle;
     __XpgpPS_iFilledRectangle();
 }
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f %.3f FRC\n",
	 height, width, y, x);
 return 0;
}


int __XpgpPS_FilledRectangleS(double x1, double y1, double x2, double y2)
{
 __XpgpPS_FilledRectangle(x1,y1,x2-x1,y2-y1);
 return 0;
}


int __XpgpPS_Arc(double x, double y,
		 double r1, double r2, double a1, double a2)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_arc)) {
     CurOD->PS.sr_flag|=XpgpPS_arc;
     __XpgpPS_iArc();
 }
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f %.3f %.3f %.3f Arc\n",
	 a1, a2, r1, r2, y, x);
 return 0;
}
 

int __XpgpPS_Circle(double x, double y, double r)
{
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f 0 360 arc stroke\n",y,x,r);
 return 0;
}
 

int __XpgpPS_Ellipse(double x, double y, double r1, double r2)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_arc)) {
     CurOD->PS.sr_flag|=XpgpPS_arc;
     __XpgpPS_iArc();
 }
 fprintf(CurOD->PS.fp,"0 360 %.3f %.3f %.3f %.3f Arc\n",
	 r1, r2, y, x);
 return 0;
}
 

int __XpgpPS_FilledArc(double x, double y,
		       double r1, double r2, double a1, double a2)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_filledArc)) {
     CurOD->PS.sr_flag|=XpgpPS_filledArc;
     __XpgpPS_iFilledArc();
 }
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f %.3f %.3f %.3f FArc\n",
	 a1, a2, r1, r2, y, x);
 return 0;
}
 

int __XpgpPS_FilledCircle(double x, double y, double r)
{
 fprintf(CurOD->PS.fp,"%.3f %.3f %.3f 0 360 arc fill\n",y,x,r);
 return 0;
}
 

int __XpgpPS_FilledEllipse(double x, double y, double r1, double r2)
{
 if(!(CurOD->PS.sr_flag & XpgpPS_filledArc)) {
     CurOD->PS.sr_flag|=XpgpPS_filledArc;
     __XpgpPS_iFilledArc();
 }
 fprintf(CurOD->PS.fp,"0 360 %.3f %.3f %.3f %.3f FArc\n",
	 r1, r2, y, x);
 return 0;
}

  

int __XpgpPS_DrawString(double x, double y, char *text)
{
 CurOD->PS.gcur.x=x;
 CurOD->PS.gcur.y=y;
 if(!(CurOD->PS.sr_flag & XpgpPS_drawString)) {
     CurOD->PS.sr_flag|=XpgpPS_drawString;
     __XpgpPS_iDrawString();
 }
 fprintf(CurOD->PS.fp,"(%s) %.3f %.3f DrawStr\n",text,y,x);
 return 0;
}


int __XpgpPS_SetQP(double d, int shape)
{
 CurOD->PS.QPd=d;
 CurOD->PS.QPshape=shape;
 return 0;
}


int __XpgpPS_PlotQP(double x, double y)
{
 int xi,yi;
 double d;
 
 d=CurOD->PS.QPd;
 xi=dtoi(x);
 yi=dtoi(y);

 switch(CurOD->PS.QPshape) {
   case 0:
     if(!(CurOD->PS.sr_flag & XpgpPS_filledRectangle)) {
	 CurOD->PS.sr_flag|=XpgpPS_filledRectangle;
	 __XpgpPS_iFilledRectangle();
     }
     fprintf(CurOD->PS.fp,"%.3f %.3f %.3f %.3f FRC\n",
	     d,d,((double)yi-0.5)*d,((double)xi-0.5)*d);
     break;
   case 1:
     fprintf(CurOD->PS.fp,"%.3f %.3f %.3f 0 360 arc fill\n",
	     yi*d,xi*d,d/2.0);
     break;
   default:
     return -1;
 }
 return 0;
}

static void put_byte(unsigned char b)
{
/*printf("%c",b);*/
fputc(b,CurOD->PS.fp);
}

unsigned char *gdata; /*global variable for get_scline */

static void get_scline(int y, int width, unsigned char *scline)
{
int i;
for(i=0;i<width;i++) scline[i] = gdata[i+y*width];
}

long RLE_encode(int,int,double,double,int,unsigned char *,unsigned char *,unsigned char*,
              unsigned char *,
              void (*)(int,int,unsigned char *), void (*)(unsigned char));

int __XpgpPS_PutImage(double x, double y,unsigned char *data,
                      unsigned int width,unsigned int height)
{
int i,j;
unsigned char R[256],G[256],B[256];
unsigned char *scline;
int NCol;

scline = (unsigned char *)malloc(width*sizeof(unsigned char));
/*printf("%s\n"," called __XpgpPS_PutImage");*/
NCol = CurOD->PS.NofInks;
if (NCol <= 256)
 {
 gdata = data;
 for(i=0;i<NCol;i++)
   {
   R[i] = (unsigned char)(CurOD->PS.inks[i].r/256);
   G[i] = (unsigned char)(CurOD->PS.inks[i].g/256);
   B[i] = (unsigned char)(CurOD->PS.inks[i].b/256);
   }
 RLE_encode(width,height,x,y,NCol,R,G,B,scline,get_scline,put_byte);
 free(scline);
 }
else
 {
  for(i=0;i<width;i++)
  for(j=0;j<height;j++)
    {
    /* set color */
    __XpgpPS_Pen(data[i+j*width]);
    /* put rectangle */
    __XpgpPS_FilledRectangle(x+(double)i,y+(double)j,1.,1.);
    }
 }
return 0;
}


int __XpgpPS_PutBitmap(double x, double y,char *bitmap,
                      unsigned int width,unsigned int height,
                      int fg,int bg)
{
return 0;
}

/*****************************************************************************

			    INITIALIZATION

 *****************************************************************************/

int __XpgpPS_iPlot()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s",
	 "/PL {\n",
	 " gsave newpath [] 0 setdash 1 setlinecap\n",
	 " 0.2 sub exch 0.2 sub exch moveto 0.4 0.4 rlineto\n",
	 " stroke grestore\n} def\n");
 return 0;
}


int __XpgpPS_iTriangle()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s",
	 "/TR {\n",
	 " newpath\n",
	 " moveto lineto lineto\n",
	 " closepath stroke\n} def\n");
 return 0;
}


int __XpgpPS_iFilledTriangle()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s",
	 "/FTR {\n",
	 " newpath\n",
	 " moveto lineto lineto\n",
	 " closepath fill\n} def\n");
 return 0;
}


int __XpgpPS_iRectangle()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s%s",
	 "/RC {\n",
	 " newpath\n",
	 " moveto dup 0 exch rlineto exch 0 rlineto\n",
	 " 0 exch sub 0 exch rlineto\n",
	 " closepath stroke\n} def\n");
 return 0;
}


int __XpgpPS_iFilledRectangle()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s%s",
	 "/FRC {\n",
	 " newpath\n",
	 " moveto dup 0 exch rlineto exch 0 rlineto\n",
	 " 0 exch sub 0 exch rlineto\n",
	 " closepath fill\n} def\n");
 return 0;
}


int __XpgpPS_iArc()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s",
	 "/Arc {\n",
	 " gsave newpath\n",
	 " translate 90 rotate scale 0 0 1 5 3 roll arc",
	 " stroke grestore \n} def\n");
 return 0;
}


int __XpgpPS_iFilledArc()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s",
	 "/FArc {\n",
	 " gsave newpath\n",
	 " translate 90 rotate scale 0 0 1 5 3 roll arc",
	 " fill grestore \n} def\n");
 return 0;
}


int __XpgpPS_iDrawString()
{
 fprintf(CurOD->PS.fp,"%s%s%s%s",
	 "/DrawStr {\n",
	 " gsave newpath\n",
	 " translate 90 rotate 0 0 moveto show\n",
	 " stroke grestore\n} def\n");
 return 0;
}


/*****************************************************************************

			    GENERAL

 *****************************************************************************/


int __XpgpPS_AllocateBasicColors(void)
{
/*
 *   "White","Black",
 *   "Light Blue","Cyan","Blue","Navy","Violet",
 *   "Magenta","Purple","Pink","Khaki","Aquamarine",
 *   "Pale Green","Green","Lime Green","Yellow Green","Green Yellow","Yellow",
 *   "Gold","Light Goldenrod","Goldenrod","Dark Goldenrod",
 *   "Brown","Chocolate","Peru","Coral",
 *   "Light Coral","Salmon",
 *   "Dark Orange","Orange","Orange Red",
 *   "Red"
 */
 int i;
 static unsigned short colors[32][3] = {
 {65535,65535,65535},{0,0,0},
 {44461,55512,59110},{0,65535,65535},{0,0,65535},{0,0,32896},{61166,33410,61166},
 {65535,0,65535},{41120,8224,61680},{65535,49344,52171},{61680,59110,35980},{32639,65535,54484},
 {39064,64507,39064},{0,65535,0},{12850,52685,12850},{39578,52685,12850},{44461,65535,12079},{65535,65535,0},
 {65535,55255,0},{61166,56797,33410},{56026,42405,8224},{47288,34438,2827},
 {42405,10794,10794},{53970,26985,7710},{52685,34181,16191},{65535,32639,20560},
 {61680,32896,32896},{64250,32896,29298},
 {65535,35980,0},{65535,42405,0},{65535,17733,0},
 {65535,0,0}};

 if(CurOD->PS.inks!=NULL)
   free(CurOD->PS.inks);
 if((CurOD->PS.inks=(struct inks *)malloc(32*sizeof(CurOD->PS.inks[0])))
    ==NULL) {
   fputs("Xpgp: not enough memory\n",stderr);
   return -1;
 }
 CurOD->PS.NofInks=32;
 for(i=0;i<32;i++) {
     CurOD->PS.inks[i].r=colors[i][0];
     CurOD->PS.inks[i].g=colors[i][1];
     CurOD->PS.inks[i].b=colors[i][2];
 }
 return 0;
}


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

int __XpgpPS_GetMouseStatus(int *x, int *y, unsigned int *status)
{
 return 0;
}

int __XpgpPS_GetMousePressEvent(int *x, int *y, unsigned int *status)
{
 return 0;
}

int __XpgpPS_fShow()
{
 return 0;
}

int __XpgpPS_SetBackground(int color)
{
 return 0;
}


int __XpgpPS_SetFocus(union __Xpgp_OutputDevice *dev)
{
  return 0;
}
